@@ -31,6 +31,6 @@ def _importer():
 cmodel, cmodel_available = _attempt_import(
     'appsi_cmodel',
     error_message=('Appsi requires building a small c++ extension. '
-                   'Please use thye "pyomo build-extensions" command'),
+                   'Please use the "pyomo build-extensions" command'),
     importer=_importer,
 )
